/*************************************************************************/
/*************************************************************************/
/* H. Timucin OZDEMIR                                                    */
/* Date : February, 1997                                                 */
/* Update : August 12, 1997                                              */
/* File Name : GAImplement.cc                                            */
/*************************************************************************/
/*************************************************************************/

#include <fstream.h>
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "FuzzyInfEngine.h"
#include "FuzzyVariable.h"
#include "FuzzySet.h"
#include "gaussMF.h"
#include "sigmoidMF.h"
#include "trapMF.h"
#include "Centroid_Area.h"
#include <StatisticSet.h>
//#include "Bisector_Area.h"
//#include "Small_Maximum.h"
//#include "Large_Maximum.h"
//#include "Mean_Maximum.h"


#include "GeneticAlgorithm.h"
#include "MutationOperator.h"
#include "CrossOverOperation.h"
#include "FitnessFunction.h"
#include "Population.h"
#include "RandomGenerator.h"
#include "deffunc.h"
#include "GAOperationsBag.h"
#include "FitnessBag.h"
#include "OperatorPicker.h"
#include "OperatorProbUpdater.h"
#include "Sampler.h"


/*
//
*/

GeneticAlgorithm::GeneticAlgorithm(FitnessFunction *_Func, 
	  CrossOverOperation *Cross, MutationOperator *Mutator,
	  ParentSelector *PSelObject, Recombinator *RecomObj, 
	  Initializer *InitObject,
	  Sampler *_MySampler, 
	  GAOperationsBag  *CrossOps, GAOperationsBag  *MutOps)
{  
   RandomGenerator r(0.0,1.0);

   Function  = _Func;

   CrossOver = Cross;
   Mutation  = Mutator;

   Random    =  r;

   ParentSelectorObject = PSelObject;
   RecombinatorObject = RecomObj;
   InitializerObject = InitObject;

   testNum = 0;
   testType = 0;

   CrossOverOprs = CrossOps;
   MutationOprs = MutOps;

   MySampler = _MySampler;

}// end of constructor

/*
GeneticAlgorithm::~GeneticAlgorithm()
{  cout << " GA is destroyed " << endl;
}// end of destructor
*/


/*
//
*/

void GeneticAlgorithm::evaluateFitness()
{  int i, limit = MyPopulation.getSize();
   Chromosomes *current;

   for(i=0;i<limit;i++)
   {  current = MyPopulation.getChromosomesIndex(i);
      (*current).setFitnessValue((*Function).apply(current));
   }// end of for

}// end of evaluateFitness()

/*
//
*/

void GeneticAlgorithm::setTestType(int a) 
{ if(a>0&&a<4) testType=a; }

/*
//
//  Collects the fitness value of the each operators result and find its
//  average and finds the average of averages so that we can say something
//  about each operator.
//  Then, if the operator's average productivity is better than the 
//  average of the all the operator's production, then reward this 
//  operator. Otherwise, punish this operator.
//
*/

void GeneticAlgorithm::updateOperatorProbabilities(GAOperationsBag& OprsBag, 
						   const Population &MyNewPop,
						   OP_TYPE operType)
{  unsigned numOfOprs = OprsBag.getSize();
   unsigned i;
   FitnessBag *bg = new FitnessBag[numOfOprs];
   Chromosomes *curr;
   unsigned popSize = MyNewPop.getSize();

   for(i=0;i<popSize;i++)
   {  curr = MyNewPop.getChromosomesIndex(i);
      if(curr != NULL )
	 bg[curr->getOperatorNumber(operType)].addNewValue(curr->getFitnessValue());
   }// end of for

   StatisticSet<FitnessBag> statFitBag;

   for(i=0;i<numOfOprs;i++) statFitBag.addElement(bg[i]);

   OperatorProbUpdater opProbUpdate(&OprsBag);
   opProbUpdate.update(statFitBag);

   delete [] bg;

}// end of updateOperatorProbabilities()

/*
//
// Until the new Population has enough chromosomes in it
//
//   1. Select the parent that undergoes to the Genetic
//      Opertaion.
//
//   2. Flip the coin for Crossover operation. If it
//      succeedes, then apply Crossover.
//
//   3. Flip the coin for Mutation Operation. If it 
//      succeedes, then apply Mutation.
//
*/

Population* GeneticAlgorithm::runOnceOn(Population &MyPopulation)
{  int i, limit = MyPopulation.getSize();
   Population *MyNewPop = new Population(limit);
   Chromosomes *Mate;
   ChromosomesBag ChroBag;
   Chromosomes *Mutant;
   Chromosomes *current;
   unsigned opick;
   CrossOverOperation *ops ;
   MutationOperator   *opsM ;

   Population SamplePopulation(MyPopulation.getSize());

   if(MySampler!=NULL)
   {
       SamplePopulation = MySampler->filterSample(MyPopulation);
       *MyNewPop        = MySampler->filterDying(MyPopulation);
   } 
   else SamplePopulation = MyPopulation;

   cout << " in GeneticAlgorithm::runEpoch2() " << endl;

   limit = maximum(limit,1);

   while( MyNewPop->getIndex() < limit )
   {   
       current = SamplePopulation.applyTournamentSelection();

       if(Random.generate()<CrossoverProbability&&
	  (CrossOver!=NULL||CrossOverOprs!=NULL))
       {  
	  cout << " We need crossover " << endl;

	  Mate = (*ParentSelectorObject).select(SamplePopulation,current);

	  if(Mate!=NULL)
          {  
	     cout << " We need to pick Crossover operation " << endl;

             ops = CrossOver;

	     if(CrossOverOprs!=NULL)
	     {  
		OperatorPicker opick(CrossOverOprs);
		ops = (CrossOverOperation*)opick.randomPick();

                if(ops==NULL) ops = CrossOver;

	     }// end of if

	     cout << " We picked Crossover operation " << endl;
	     // In any case, if is comes empty guarantee that we
	     // have something to apply 

	     if(ops==NULL) ops = CrossOver;

	     cout << " Before we apply crossover " << endl;
	     ChroBag = ops->apply(current,Mate,CrossoverProbability);
	     cout << " After we apply crossover " << endl;

	     /* FitnessValues should be updated here */

             MyNewPop->addChromosomesBag(ChroBag);

	     cout << " After we added the result to the Population " << endl;

          } else cout << " We could not find any mate " << endl;

       }// end of if

       if(Random.generate()<MutationProbability &&
	  (Mutation!=NULL||MutationOprs!=NULL))
       {  
	  cout << " We need mutation " << endl;

	  Mate = (*ParentSelectorObject).select(SamplePopulation,current);

	  if(Mate!=NULL)
	  {
             opsM = Mutation;

	     if(MutationOprs!=NULL)
	     {  
                OperatorPicker opick(MutationOprs);
	        opsM = (MutationOperator*)opick.randomPick();

                if(opsM==NULL) opsM = Mutation;

	     }// end of if

	     cout << " before  mutation.apply()  " << endl;
	     ChroBag = opsM->apply(current,Mate,MutationProbability);
	     //Mutant = opsM->apply(current);

	     cout << " after  mutation.apply()  " << endl;

	     //(*Mutant).setFitnessValue((*Function).apply(Mutant));

             MyNewPop->addChromosomesBag(ChroBag);
	     //MyNewPop->addChromosomes(Mutant);

	     cout << " after MyNewPop.addChromosomes " << endl;

	  }// end of if

       }// end of if 

   }// end of for

   cout << " End of runEpoch2() " << endl;

   return MyNewPop;

}// end of runOnceOn()

/*
//
// Until the new Population has enough chromosomes in it
//
//   1. Select the parent that undergoes to the Genetic
//      Opertaion.
//
//   2. Flip the coin for Crossover operation. If it
//      succeedes, then apply Crossover.
//
//   3. Flip the coin for Mutation Operation. If it 
//      succeedes, then apply Mutation.
//
*/

void GeneticAlgorithm::runEpoch2(Population &MyNewPop)
{  int i, limit = MyPopulation.getSize();
   Chromosomes *Mate;
   ChromosomesBag ChroBag;
   Chromosomes *Mutant;
   Chromosomes *current;
   unsigned opick;
   CrossOverOperation *ops ;
   MutationOperator   *opsM ;

   Population SamplePopulation(MyPopulation.getSize());

   if(MySampler!=NULL)
   {
       SamplePopulation = MySampler->filterSample(MyPopulation);
       MyNewPop         = MySampler->filterDying(MyPopulation);
   } 
   else SamplePopulation = MyPopulation;

   cout << " in GeneticAlgorithm::runEpoch2() " << endl;

   limit = maximum(limit-1,1);

   while( MyNewPop.getSize() < limit )
   {   
       current = SamplePopulation.applyTournamentSelection();

       if(Random.generate()<CrossoverProbability&&
	  (CrossOver!=NULL||CrossOverOprs!=NULL))
       {  
	  cout << " We need crossover " << endl;

	  Mate = (*ParentSelectorObject).select(SamplePopulation,current);

	  if(Mate)
          {  
	     cout << " We need to pick Crossover operation " << endl;

             ops = CrossOver;

	     if(CrossOverOprs!=NULL)
	     {  
		OperatorPicker opick(CrossOverOprs);
		ops = (CrossOverOperation*)opick.randomPick();

                if(ops==NULL) ops = CrossOver;

	     }// end of if

	     cout << " We picked Crossover operation " << endl;
	     // In any case, if is comes empty guarantee that we
	     // have something to apply 

	     if(ops==NULL) ops = CrossOver;

	     cout << " Before we apply crossover " << endl;
	     ChroBag = ops->apply(current,Mate);
	     cout << " After we apply crossover " << endl;

	     /* FitnessValues should be updated here */

             MyNewPop.addChromosomesBag(ChroBag);

	     cout << " After we added the result to the Population " << endl;

          } else cout << " We could not find any mate " << endl;

       }// end of if

       if(Random.generate()<MutationProbability &&
	  (Mutation!=NULL||MutationOprs!=NULL))
       {  
	  cout << " We need mutation " << endl;

          opsM = Mutation;

	  if(MutationOprs!=NULL)
	  {  
             OperatorPicker opick(MutationOprs);
	     opsM = (MutationOperator*)opick.randomPick();

             if(opsM==NULL) opsM = Mutation;

	  }// end of if

	  Mutant = opsM->apply(current);

	  cout << " after  mutation.apply()  " << endl;

	  //(*Mutant).setFitnessValue((*Function).apply(Mutant));

	  cout << " after setFitnessValue " << endl;

	  MyNewPop.addChromosomes(Mutant);

	  cout << " after MyNewPop.addChromosomes " << endl;

       }// end of if 

   }// end of for

   cout << " End of runEpoch2() " << endl;

}// end of runEpoch2()

/*
//
*/

void GeneticAlgorithm::runEpoch(Population &MyNewPop)
{  int i, limit = MyPopulation.getSize();
   Chromosomes *Mate;
   ChromosomesBag ChroBag;
   Chromosomes *Mutant;
   Chromosomes *current;
   unsigned opick;
   CrossOverOperation *ops ;
   MutationOperator   *opsM ;

   cout << " in GeneticAlgorithm::runEpoch() " << endl;

   for(i=0;i<limit;i++)
   {   
       current = MyPopulation.getChromosomesIndex(i);

       if(Random.generate()<CrossoverProbability&&
	  (CrossOver!=NULL||CrossOverOprs!=NULL))
       {  
	  cout << " We need crossover " << endl;

	  Mate = (*ParentSelectorObject).select(MyPopulation,current);

	  if(Mate)
          {  
	     cout << " We need to pick Crossover operation " << endl;

             ops = CrossOver;

	     if(CrossOverOprs!=NULL)
	     {  
                OperatorPicker opick(CrossOverOprs);
	        ops = (CrossOverOperation*) opick.randomPick();

                if(ops==NULL) ops = CrossOver;

	     }// end of if

	     cout << " We picked Crossover operation " << endl;
	     // In any case, if is comes empty guarantee that we
	     // have something to apply 

	     if(ops==NULL) ops = CrossOver;

	     cout << " Before we apply crossover " << endl;
	     ChroBag = ops->apply(current,Mate);
	     cout << " After we apply crossover " << endl;

	     /* FitnessValues should be updated here */

             MyNewPop.addChromosomesBag(ChroBag);

	     cout << " After we added the result to the Population " << endl;

          } else cout << " We could not find any mate " << endl;

       }// end of if

       if(Random.generate()<MutationProbability &&
	  (Mutation!=NULL||MutationOprs!=NULL))
       {  
	  cout << " We need mutation " << endl;

          opsM = Mutation;

	  if(MutationOprs!=NULL)
	  {  
             OperatorPicker opick(MutationOprs);
	     opsM = (MutationOperator*)opick.randomPick();

             if(opsM==NULL) opsM = Mutation;

	  }// end of if

	  Mutant = opsM->apply(current);

	  cout << " after  mutation.apply()  " << endl;

	  //(*Mutant).setFitnessValue((*Function).apply(Mutant));

	  cout << " after setFitnessValue " << endl;

	  MyNewPop.addChromosomes(Mutant);

	  cout << " after MyNewPop.addChromosomes " << endl;

       }// end of if 

   }// end of for

   cout << " End of runEpoch() " << endl;

}// end of runEpoch()

/*
//
*/

void GeneticAlgorithm::run()
{  int i=0;
   Population *NewPop;
   Population temp(PopulationSize);

   MyPopulation = temp;

   //MyPopulation.setPopulationSize(PopulationSize);

   cout << " (*InitializerObject).apply " << endl;
   (*InitializerObject).apply(MyPopulation);

   evaluateFitness();

   char fileName[30];
   sprintf(fileName,"GAStatisticsT%d.out",testNum+1);

   ofstream oStat(fileName); oStat.close();

   i=0;

   cout << " Initial Population is " << MyPopulation << endl;
   cout << " Initial Population is printed! " << endl;

   while(i<MaxGeneration)
   {  
      cout << " Before definition of new population " << endl;
      NewPop = new Population(0);
      cout << " After definition of new population " << endl;

      writeStatistics(i, fileName);
      cout << " After writeStatistics() " << endl;

      runEpoch(*NewPop);

      cout << " Combine the Population " << endl;

      (*NewPop).printFitnessValues(cout);

      cout << " Current Population is " << MyPopulation << endl;
      cout << " New Generation is " << *NewPop << endl;

      MyPopulation = (*RecombinatorObject).apply(MyPopulation,*NewPop);

      cout << " Combine is successfull " << endl;

      MyPopulation.printFitnessValues(cout);

      cout << " After printFitnessValues() " << endl;

      i++;

      if(CrossOverOprs!=NULL)
             updateOperatorProbabilities(*CrossOverOprs,MyPopulation,CROSSOVER);

      if(MutationOprs!=NULL)
             updateOperatorProbabilities(*MutationOprs,MyPopulation,MUTATION);

      cout << endl << " run " << i << endl ;//<< MyPopulation;

      
   }// end of while()

   MyPopulation.printFitnessValues(cout);

   writeStatistics(i, fileName);

   cout << " The result is" << endl;
   cout << MyPopulation << endl;

   sprintf(fileName,"GASolutionT%d.out",testNum+1);
   ofstream ofile(fileName);
   if(ofile)
   {  
      MyPopulation.printFitnessValues(ofile);
      ofile << MyPopulation << endl;
      ofile.close();
   }// end of if

}// end of run()

/*
//
*/

void GeneticAlgorithm::runFuzzy()
{  sigmoidMF none(0.05,-100.0);
   trapMF    small(0.05,0.08,0.09,0.13);
   trapMF    large(0.10,0.18,0.25,0.35);
   sigmoidMF VLarge(0.4,25.0);
   
   //write_func(&none,"none.M");
   //write_func(&small,"small.M");
   //write_func(&large,"large.M");
   //write_func(&VLarge,"vlarge.M");

   sigmoidMF noneS(0.1,-100.0);
   //trapMF    smallS(0.09,0.18,0.23,0.29);
   //trapMF    largeS(0.23,0.35,0.42,0.49);
   //sigmoidMF VLargeS(0.5,25.0);
   trapMF    smallS(0.19,0.28,0.33,0.49);
   trapMF    largeS(0.33,0.45,0.53,0.69);
   sigmoidMF VLargeS(0.65,25.0);

   //write_func(&noneS,"noneS.M");
   //write_func(&smallS,"smallS.M");
   //write_func(&largeS,"largeS.M");
   //write_func(&VLargeS,"vlargeS.M");

   gaussMF decrease(0.2,0.2);
   gaussMF stay(0.5,0.1);
   gaussMF increase(0.8,0.2);

   FuzzySet st(&none);
   FuzzyVariable NumExactMatches(st);
   st.setMFunction(&small);
   NumExactMatches.addClass(st);
   st.setMFunction(&large);
   NumExactMatches.addClass(st);
   st.setMFunction(&VLarge);
   NumExactMatches.addClass(st);

   st.setMFunction(&noneS);
   FuzzyVariable SofD(st);
   st.setMFunction(&smallS);
   SofD.addClass(st);
   st.setMFunction(&largeS);
   SofD.addClass(st);
   st.setMFunction(&VLargeS);
   SofD.addClass(st);

   st.setMFunction(&decrease);
   FuzzyVariable CrossOperation(st);
   st.setMFunction(&stay);
   CrossOperation.addClass(st);
   st.setMFunction(&increase);
   CrossOperation.addClass(st);

   st.setMFunction(&decrease);
   FuzzyVariable MutaOperation(st);
   st.setMFunction(&stay);
   MutaOperation.addClass(st);
   st.setMFunction(&increase);
   MutaOperation.addClass(st);

   FuzzyInfEngine myEngine(2,2);
   myEngine.addInput(&NumExactMatches);
   myEngine.addInput(&SofD);

   myEngine.addOutput(&CrossOperation);
   myEngine.addOutput(&MutaOperation);

   float inrule1[2] = { 1.0, 1.0};
   float inrule2[2] = { 1.0, 2.0};
   float inrule3[2] = { 1.0, 3.0};
   float inrule4[2] = { 1.0, 4.0};

   float outrule1[2] = { 1.0, 1.0};
   float outrule2[2] = { 1.0, 2.0};
   float outrule3[2] = { 1.0, 3.0};
   float outrule4[2] = { 2.0, 1.0};
   float outrule5[2] = { 2.0, 2.0};
   float outrule6[2] = { 2.0, 3.0};
   float outrule7[2] = { 3.0, 3.0};

   /*  Previous Rules
   Vector inp1(inrule1,2);
   Vector out1(outrule4,2);
   myEngine.addRule(inp1,out1);

   Vector inp2(inrule2,2);
   Vector out2(outrule4,2);
   myEngine.addRule(inp2,out2);

   Vector inp3(inrule3,2);
   Vector out3(outrule1,2);
   myEngine.addRule(inp3,out3);

   Vector inp4(inrule4,2);
   Vector out4(outrule2,2);
   myEngine.addRule(inp4,out4);
   */

   Vector inp1(inrule1,2);
   Vector out1(outrule6,2);
   myEngine.addRule(inp1,out1);

   Vector inp2(inrule2,2);
   Vector out2(outrule5,2);
   myEngine.addRule(inp2,out2);

   Vector inp3(inrule3,2);
   Vector out3(outrule5,2);
   myEngine.addRule(inp3,out3);

   Vector inp4(inrule4,2);
   Vector out4(outrule1,2);
   myEngine.addRule(inp4,out4);

   float inrule5[2] = { 2.0, 1.0};
   float inrule6[2] = { 2.0, 2.0};
   float inrule7[2] = { 2.0, 3.0};
   float inrule8[2] = { 2.0, 4.0};

   Vector inp5(inrule5,2);
   Vector out5(outrule3,2);
   myEngine.addRule(inp5,out5);

   Vector inp6(inrule6,2);
   Vector out6(outrule5,2);
   myEngine.addRule(inp6,out6);

   Vector inp7(inrule7,2);
   Vector out7(outrule5,2);
   myEngine.addRule(inp7,out7);

   Vector inp8(inrule8,2);
   Vector out8(outrule5,2);
   myEngine.addRule(inp8,out8);

   float inrule9[2] = { 3.0, 1.0};
   float inrule10[2] = { 3.0, 2.0};
   float inrule11[2] = { 3.0, 3.0};
   float inrule12[2] = { 3.0, 4.0};

   Vector inp9(inrule9,2);
   Vector out9(outrule7,2);
   myEngine.addRule(inp9,out9);

   Vector inp10(inrule10,2);
   Vector out10(outrule7,2);
   myEngine.addRule(inp10,out10);

   Vector inp11(inrule11,2);
   Vector out11(outrule6,2);
   myEngine.addRule(inp11,out11);

   Vector inp12(inrule12,2);
   Vector out12(outrule5,2);
   myEngine.addRule(inp12,out12);

   float inrule13[2] = { 4.0, 1.0};
   float inrule14[2] = { 4.0, 2.0};
   float inrule15[2] = { 4.0, 3.0};
   float inrule16[2] = { 4.0, 4.0};

   Vector inp13(inrule13,2);
   Vector out13(outrule7,2);
   myEngine.addRule(inp13,out13);

   Vector inp14(inrule14,2);
   Vector out14(outrule7,2);
   myEngine.addRule(inp14,out14);

   Vector inp15(inrule15,2);
   Vector out15(outrule7,2);
   myEngine.addRule(inp15,out15);

   Vector inp16(inrule16,2);
   Vector out16(outrule6,2);
   myEngine.addRule(inp16,out16);

   cout << " My engine has " << myEngine << endl;


   /*float test11[2] = { 0.01, 0.1};
   float test12[2] = { 0.01, 0.2};
   float test13[2] = { 0.01, 0.3};
   float test14[2] = { 0.01, 0.6};

   float test21[2] = { 0.1, 0.1};
   float test22[2] = { 0.1, 0.2};
   float test23[2] = { 0.1, 0.3};
   float test24[2] = { 0.1, 0.6};

   float test31[2] = { 0.22, 0.1};
   float test32[2] = { 0.22, 0.2};
   float test33[2] = { 0.22, 0.3};
   float test34[2] = { 0.22, 0.6};

   float test41[2] = { 0.6, 0.1};
   float test42[2] = { 0.6, 0.2};
   float test43[2] = { 0.6, 0.3};
   float test44[2] = { 0.6, 0.6};

   Vector **tests = new Vector*[16];

   tests[0] = new Vector(test11,2);
   tests[1] = new Vector(test12,2);
   tests[2] = new Vector(test13,2);
   tests[3] = new Vector(test14,2);

   tests[4] = new Vector(test21,2);
   tests[5] = new Vector(test22,2);
   tests[6] = new Vector(test23,2);
   tests[7] = new Vector(test24,2);

   tests[8] = new Vector(test31,2);
   tests[9] = new Vector(test32,2);
   tests[10] = new Vector(test33,2);
   tests[11] = new Vector(test34,2);

   tests[12] = new Vector(test41,2);
   tests[13] = new Vector(test42,2);
   tests[14] = new Vector(test43,2);
   tests[15] = new Vector(test44,2);

   Centroid_Area def; 
  //Bisector_Area def(100); 
  //Small_Maximum def(100); 
  //Large_Maximum def(100); 
  //Mean_Maximum def(100); 
   Vector res(test41,2);

   for(int i=0;i<16;i++)
   {  cout << " let's test " << *tests[i] << endl; 
      res = myEngine.RunEngine(*tests[i],&def);
      
      cout << i+1 << " test " << res << endl;

      cout << " decrease " << decrease.getValue(res.getElement(0)) << endl;
      cout << " keep " << stay.getValue(res.getElement(0)) << endl;
      cout << " increase " << increase.getValue(res.getElement(0)) << endl;

      cout <<  *tests[i] << " action CROSSOVER is " ;
      switch ( CrossOperation.getClassNumber(res.getElement(0)) )
      {  case 1 : cout << " DECREASE " << endl;
		  break;
         case 2 : cout << " KEEP " << endl;
		  break;
         case 3 : cout << " INCREASE " << endl;
		  break;
         default : cout << " default " << endl;
		  break;
      }// end of switch

      cout <<  *tests[i] << " action for MUTATION is " ;
      switch ( MutaOperation.getClassNumber(res.getElement(1)) )
      {  case 1 : cout << " DECREASE " << endl;
		  break;
         case 2 : cout << " KEEP " << endl;
		  break;
         case 3 : cout << " INCREASE " << endl;
		  break;
         default : cout << " default " << endl;
		  break;
      }// end of switch

   }// end of for
   */

/*
   cout << (*tests[0]).getElement(0) << " none : " << none.getValue((*tests[0]).getElement(0)) << endl;
   cout << " small : " << small.getValue((*tests[0]).getElement(0)) << endl;
   cout << " large : " << large.getValue((*tests[0]).getElement(0)) << endl;
   cout << " Vlarge : " << VLarge.getValue((*tests[0]).getElement(0)) << endl;

   cout << (*tests[0]).getElement(1) << " noneS : " << noneS.getValue((*tests[0]).getElement(1)) << endl;
   cout << " smallS : " << smallS.getValue((*tests[0]).getElement(1)) << endl;
   cout << " largeS : " << largeS.getValue((*tests[0]).getElement(1)) << endl;
   cout << " VlargeS : " << VLargeS.getValue((*tests[0]).getElement(1)) << endl;
*/

   Centroid_Area def; 
   Vector res(inrule9,2);
   int i=0;
   Population *NewPop;
   Population temp(PopulationSize);

   MyPopulation = temp;

   //MyPopulation.setPopulationSize(PopulationSize);

   cout << " (*InitializerObject).apply " << endl;
   (*InitializerObject).apply(MyPopulation);

   evaluateFitness();

   i=0;

   char fileName[30];

   sprintf(fileName,"GAStatisticsT%d.out",testNum+1);

   //* Refresh the output file

   ofstream oStat(fileName); oStat.close();

   while(i<MaxGeneration)
   {  
      NewPop = new Population(0);

      writeStatistics(i, fileName);

      runEpoch(*NewPop);

      cout << " Combine the Population " << endl;

      (*NewPop).printFitnessValues(cout);

      MyPopulation = (*RecombinatorObject).apply(MyPopulation,*NewPop);

      cout << " Combine is successfull " << endl;

      MyPopulation.printFitnessValues(cout);
      i++;

      cout << endl << " run " << i << endl ;//<< MyPopulation;

      if(testType)
      {
      cout << " NumberOfMatches : " << MyPopulation.getNumberOfMatches() << endl;
      cout << " SumOfDissimilarities : " << MyPopulation.getSumOfDissimilarities() << endl;

      float inP[2];
      inP[0] = MyPopulation.getNumberOfMatches();
      inP[1] = MyPopulation.getSumOfDissimilarities();
      Vector inVe(inP,2);
      res = myEngine.RunEngine(inVe,&def);

      if(testType==1 || testType==3)
      {
         cout <<  inVe << " action CROSSOVER is " ;
         switch ( CrossOperation.getClassNumber(res.getElement(0)) )
         {  case 1 : cout << " DECREASE " << endl;
		     CrossoverProbability = maximum(0.9*CrossoverProbability,0.0);
		     break;
            case 2 : cout << " KEEP " << endl;
		     break;
            case 3 : cout << " INCREASE " << endl;
		     CrossoverProbability = minimum(1.1*CrossoverProbability,0.8);
		     break;
           default : cout << " default " << endl;
		     break;
         }// end of switch

      }// end of if

      if(testType==2 || testType==3)
      {  cout <<  inVe << " action for MUTATION is " ;
         switch ( MutaOperation.getClassNumber(res.getElement(1)) )
         {  case 1 : cout << " DECREASE " << endl;
	             MutationProbability = maximum(0.9 * MutationProbability,0.0);
		     break;
            case 2 : cout << " KEEP " << endl;
		     break;
            case 3 : cout << " INCREASE " << endl;
		     MutationProbability = minimum(1.1 * MutationProbability,0.3);
		     break;
           default : cout << " default " << endl;
		     break;
        }// end of switch

      }// end of if

      }// end of if

   }// end of while()

   writeStatistics(i, fileName);

   MyPopulation.printFitnessValues(cout);

   cout << " The result is" << endl;
   cout << MyPopulation << endl;

   sprintf(fileName,"GASolutionT%d.out",testNum+1);
   ofstream ofile(fileName);
   if(ofile)
   {  MyPopulation.printFitnessValues(ofile);
      ofile << MyPopulation << endl;
      ofile.close();
   }// end of if

}// end of runFuzzy()

/*
//  Runs the GA m times
*/

void GeneticAlgorithm::runFuzzyTimes(int m)
{  int i=0;
   float Pcross = CrossoverProbability;
   float Pmutat = MutationProbability;
   time_t start;
   time_t stop;
   char fileName[30];


   while(i<m)
   {  testNum=i;
      start = time(NULL);
      cout << " \n start " << ctime(&start) << endl;
      runFuzzy();
      stop = time(NULL);
      cout << " \n stop " << ctime(&stop) << endl;
      sprintf(fileName,"GASolutionT%d.out",i+1);
      ofstream ofile(fileName,ios::app);

      if(ofile)
      {   ofile << endl << endl ; 
	  ofile << " The calculatation took " << difftime(stop,start);
	  ofile << " secs \n\n" << endl;
	  ofile.close();
      }

      CrossoverProbability = Pcross;
      MutationProbability = Pmutat;
      MyPopulation.clearBag();
      i++;
   }// end of while

}// end of runFuzzyTimes()

/*
//
*/

void GeneticAlgorithm::SetMaxGeneration(int _m)
{  MaxGeneration=_m;
   cout << endl << " Maximum Generation Number is " << MaxGeneration << endl;
}// end of SetMaxGeneration()

/*
//
*/

void GeneticAlgorithm::SetPopulationSize(int _size)
{  PopulationSize=_size;
   cout << endl << " Population Size is " << PopulationSize << endl;
}// end of SetPopulationSize()

/*
//
*/

void GeneticAlgorithm::SetMutationProb(float _ProbMut)
{  MutationProbability=_ProbMut;
   cout << endl << " Mutation Probability is " << MutationProbability << endl;
}// end of SetMutationProb()

/*
//
*/

void GeneticAlgorithm::SetCrossProb(float _ProbCross)
{  CrossoverProbability=_ProbCross;
   cout << endl << " Crossover Probability is " << CrossoverProbability << endl;
}// end of SetCrossProb()

/*
//
*/

void GeneticAlgorithm::writeStatistics(int i, char *FileName)
{
   FitnessValue *T = NULL;
   ofstream ofile(FileName,ios::app);

   cout << " In writeStatistics() " << endl;

   //* Write statistics about the GA

   ofile << i << " "; // write generation number

   T = MyPopulation.getWorstFitness();
   if( T != NULL ) ofile <<  *T << " "; 

   T = MyPopulation.getBestFitness();
   if( T != NULL ) ofile <<  *T << " "; 

   T = MyPopulation.getAverageFitness();
   if( T != NULL ) ofile <<  *T << " "; 

   //T = MyPopulation.getVarianceFitness();
   //if( T != NULL ) ofile <<  *T << " "; 

   ofile <<  CrossoverProbability  << " "; 
   ofile <<  MutationProbability  << " "; 
   ofile <<  MyPopulation.getNumberOfMatches()  << " "; 
   ofile <<  MyPopulation.getSumOfDissimilarities()  << endl; 
   ofile.close();

   cout << " I am leaving writeStatistics() " << endl;

}// end of writeStatics()

